import logging
import os
import re
import sys

import yaml
from mkdocs.config import config_options
from mkdocs.plugins import BasePlugin
from mkdocs.structure.files import get_files

import mkapi
import mkapi.plugins.preprocess
from mkapi.core.page import Page

logger = logging.getLogger("mkdocs")


class MkapiPlugin(BasePlugin):
    config_scheme = (
        ("src_dirs", config_options.Type(list, default=[])),
    )

    def on_config(self, config):
        self.pages = {}
        # api_dir = os.path.join(config["docs_dir"], self.config["api_dir"])
        dirname = os.path.dirname(config["config_file_path"])
        nav = config["nav"]
        for page in nav:
            for key, value in page.items():
                if isinstance(value, str) and value.startswith("mkapi:"):
                    root = value.split(":")[1]
                    root = os.path.join(dirname, root)
                    # contents = mkapi.plugins.preprocess.make_pages(root, api_dir)
                    # page[key] = contents
        for src_dir in self.config["src_dirs"]:
            if src_dir not in sys.path:
                sys.path.insert(0, src_dir)
        return config

    def on_files(self, files, config):
        root = os.path.join(os.path.dirname(mkapi.__file__), "theme")
        docs_dir = config["docs_dir"]
        config["docs_dir"] = root
        files_ = get_files(config)
        config["docs_dir"] = docs_dir

        css = []
        js = []
        for file in files_:
            path = file.src_path.replace("\\", "/")
            if path.endswith(".css"):
                files.append(file)
                css.append(path)
            elif path.endswith(".js"):
                files.append(file)
                js.append(path)
            elif path.endswith(".yml"):
                path = os.path.normpath(os.path.join(root, path))
                with open(path) as f:
                    data = yaml.safe_load(f)
                css = data.get("extra_css", []) + css
                js = data.get("extra_javascript", []) + js
        css = [x for x in css if x not in config["extra_css"]]
        js = [x for x in js if x not in config["extra_javascript"]]
        config["extra_css"] = css + config["extra_css"]
        config["extra_javascript"] = js + config["extra_javascript"]

        for file in files:
            normalize_file(file, config)

        return files

    def on_page_markdown(self, markdown, page, config, files):
        path = page.file.abs_src_path
        page = Page(markdown)
        self.pages[path] = page
        return page.markdown

    def on_page_content(self, html, page, config, files):
        path = page.file.abs_src_path
        page = self.pages[path]
        return page.content(html)

    def on_serve(self, server, config, builder):
        for path in ["theme", "templates"]:
            root = os.path.join(os.path.dirname(mkapi.__file__), path)
            server.watch(root, builder)
        return server


NORMALIZE_PATTERN = re.compile(r"(^|[\\/])\w*[0-9]+[._ ](.*?)")


def normalize_file(file, config):
    if file.dest_path.endswith(".html"):
        file.dest_path = NORMALIZE_PATTERN.sub(r"\1\2", file.dest_path)
        file.dest_path = file.dest_path.replace(" ", "_")
        file.abs_dest_path = os.path.normpath(
            os.path.join(config["site_dir"], file.dest_path)
        )
        file.url = NORMALIZE_PATTERN.sub(r"\1\2", file.url)
        file.url = file.url.replace(" ", "_")



import os
from dataclasses import dataclass, field
from typing import Dict, List

from jinja2 import Environment, FileSystemLoader, Template, select_autoescape

import mkapi


@dataclass
class Renderer:
    templates: Dict[str, Template] = field(default_factory=dict, init=False)

    def __post_init__(self):
        path = os.path.join(os.path.dirname(mkapi.__file__), "templates")
        loader = FileSystemLoader(path)
        env = Environment(loader=loader, autoescape=select_autoescape(["jinja2"]))
        for name in os.listdir(path):
            template = env.get_template(name)
            name = os.path.splitext(name)[0]
            self.templates[name] = template

    def render(self, node, parent=None) -> str:
        docstring = self.render_docstring(node.docstring)
        members = []
        if node.members:
            members = [self.render(member, node) for member in node.members]
        return self.render_node(node, docstring, members, parent)

    def render_node(self, node, docstring: str, members: List[str], parent) -> str:
        template = self.templates["node"]
        return template.render(
            node=node, docstring=docstring, members=members, parent=parent
        )

    def render_docstring(self, docstring) -> str:
        template = self.templates["docstring"]
        for section in docstring.sections:
            if section.items:
                section.html = self.render_section(section)
        return template.render(docstring=docstring)

    def render_code(self, code: str) -> str:
        template = self.templates["code"]
        return template.render(code=code)

    def render_section(self, section) -> str:
        if section.name in ["Parameters", "Attributes", "Raises"]:
            return self.templates["args"].render(section=section)
        else:
            raise ValueError(f"Invalid section name: {section.name}")



import inspect
import re


class Signature:
    def __init__(self, obj):
        self.signature = inspect.signature(obj)

    def __str__(self):
        s = str(self.signature)
        s = re.sub(r"\[[^:->]*\]", "", s)
        s = re.sub(r":.*?(,|= |\))", r"\1", s)
        s = s.replace("= ", "=")
        s = re.sub(r" ->.*", r"", s)
        s = re.sub(r"(self|cls)(, |)", "", s)
        return s


class Annotation:
    def __init__(self, obj):
        signature = inspect.signature(obj)
        self.parameters = signature.parameters
        self.defaults = {}
        for name, param in self.parameters.items():
            if param.default == inspect.Parameter.empty:
                self.defaults[name] = param.default
            else:
                self.defaults[name] = str(param.default)
        self.return_annotation = signature.return_annotation
        if hasattr(obj, "__dataclass_fields__"):
            values = obj.__dataclass_fields__.values()
            self.attributes = {f.name: to_string(f.type) for f in values}
        else:
            self.attributes = None

    def __contains__(self, name):
        return name in self.parameters

    def __getitem__(self, name):
        type = to_string(self.parameters[name].annotation)
        if self.defaults[name] != inspect.Parameter.empty:
            type += ", optional"
        return type

    @property
    def returns(self):
        return to_string(self.return_annotation)

    @property
    def yields(self):
        return to_string(self.return_annotation, "yields")


def to_string(annotation, kind: str = "") -> str:
    if kind == "yields":
        if hasattr(annotation, "__args__") and annotation.__args__:
            return to_string(annotation.__args__[0])
        else:
            return ""

    if annotation == inspect.Parameter.empty:
        return ""
    if hasattr(annotation, "__name__"):
        return annotation.__name__
    if annotation.__origin__ == list:
        type = "list of " + to_string(annotation.__args__[0])
        if type.endswith(" of T"):
            return "list"
        return type
    if annotation.__origin__ == tuple:
        args = [to_string(x) for x in annotation.__args__]
        if args:
            return ", ".join(args)
        else:
            return "tuple"
    return str(annotation).replace("typing.", "")



import inspect
import re
from dataclasses import dataclass, field
from typing import Any, Iterator, List, Optional, Tuple

from mkapi.core.inspect import Annotation

SECTIONS = [
    "Args",
    "Arguments",
    "Attributes",
    "Example",
    "Examples",
    "Note",
    "Notes",
    "Raises",
    "Returns",
    "References",
    "See Also",
    "Todo",
    "Warning",
    "Warnings",
    "Warns",
    "Yield",
    "Yields",
]


def get_indent(line: str) -> int:
    indent = 0
    for x in line:
        if x != " ":
            return indent
        indent += 1
    return -1


def join(lines):
    indent = get_indent(lines[0])
    return "\n".join(line[indent:] for line in lines)


def rename_section(name: str) -> str:
    if name in ["Args", "Arguments"]:
        return "Parameters"
    if name == "Warns":
        return "Warnings"
    return name


def split_section(doc: str) -> Iterator[Tuple[str, str]]:
    """Yields section name and its contents."""
    lines = [x.rstrip() for x in doc.split("\n")]
    name = ""
    start = indent = 0
    for stop, line in enumerate(lines, 1):
        if stop == len(lines):
            next_indent = -1
        else:
            next_indent = get_indent(lines[stop])
        if not line and next_indent < indent:
            if start < stop - 1:
                yield name, join(lines[start : stop - 1])
            start = stop
            name = ""
        elif line.endswith(":") and line[:-1] in SECTIONS:
            if start < stop - 1:
                yield name, join(lines[start : stop - 1])
            name = rename_section(line[:-1])
            start = stop
            indent = next_indent
    if start < len(lines):
        yield name, join(lines[start:])


def split_parameter(doc: str) -> Iterator[List[str]]:
    """Yields list of parameter string."""
    lines = [x.rstrip() for x in doc.split("\n")]
    start = stop = 0
    for stop, line in enumerate(lines, 1):
        if stop == len(lines):
            next_indent = 0
        else:
            next_indent = get_indent(lines[stop])
        if next_indent == 0:
            yield lines[start:stop]
            start = stop


def parse_parameter(lines: List[str]) -> Tuple[str, str, str]:
    """Yields (name, type, markdown)."""
    name, _, line = lines[0].partition(":")
    name = name.strip()
    line = line.strip()
    parsed = [line]
    if len(lines) > 1:
        indent = get_indent(lines[1])
        for line in lines[1:]:
            parsed.append(line[indent:])
    m = re.match(r"(.*?)\s*?\((.*?)\)", name)
    if m:
        name, type = m.group(1), m.group(2)
    else:
        type = ""
    return name, type, "\n".join(parsed)


def parse_parameters(doc: str) -> List[Tuple[str, str, str]]:
    """Returns list of (name, type, markdown)."""
    return [parse_parameter(lines) for lines in split_parameter(doc)]


def parse_returns(doc: str) -> Tuple[str, str]:
    """Returns (type, markdown)."""
    lines = doc.split("\n")
    if ":" in lines[0]:
        type, _, lines[0] = lines[0].partition(":")
        type = type.strip()
        lines[0] = lines[0].strip()
    else:
        type = ""
    return type, "\n".join(lines)


def parse_raise(lines: List[str]) -> Tuple[str, str]:
    """Returns (type, markdown)."""
    type, _, markdown = parse_parameter(lines)
    return type, markdown


def parse_raises(doc: str) -> List[Tuple[str, str]]:
    return [parse_raise(lines) for lines in split_parameter(doc)]


@dataclass
class Item:
    name: str
    type: str
    markdown: str
    html: str = ""

    def set_html(self, html):
        html = html.replace("<p>", "").replace("</p>", "<br>")
        if html.endswith("<br>"):
            html = html[:-4]
        self.html = html


@dataclass
class Section:
    name: str
    type: str = ""
    markdown: str = ""
    items: List[Item] = field(default_factory=list)
    html: str = ""

    def __getitem__(self, index):
        return self.items[index]

    def __len__(self):
        return len(self.items)

    def __getattr__(self, name):
        for item in self.items:
            if item.name == name:
                return item

    def __iter__(self):
        if self.markdown:
            yield self
        else:
            yield from self.items

    def set_html(self, html):
        self.html = html


@dataclass
class Docstring:
    obj: Optional[Any] = field(repr=False)
    sections: List[Section]

    def __getattr__(self, name):
        for section in self.sections:
            if section.name == name:
                return section

    def __getitem__(self, name):
        return getattr(self, name)

    def __len__(self):
        return len(self.sections)

    def __iter__(self):
        for section in self.sections:
            yield from section


def create_section(name: str, doc: str) -> Section:
    type = ""
    markdown = ""
    items = []
    if name in ["Parameters", "Attributes"]:
        items = [Item(n, t, m) for n, t, m in parse_parameters(doc)]
    elif name == "Raises":
        items = [Item(t, "", m) for t, m in parse_raises(doc)]
    elif name in ["Returns", "Yields"]:
        type, markdown = parse_returns(doc)
    else:
        markdown = doc
    return Section(name, type=type, markdown=markdown, items=items)


def postprocess(doc: Docstring, obj: Any):
    if isinstance(obj, property):
        section = doc.sections[0]
        markdown = section.markdown
        line = markdown.split("\n")[0]
        if ":" in line:
            index = line.index(":")
            section.type = line[:index].strip()
            section.markdown = markdown[index + 1 :].strip()

    if not callable(obj):
        return

    annotation = Annotation(obj)
    if doc["Parameters"] is not None:
        for item in doc["Parameters"]:
            if item.type == "" and item.name in annotation:
                item.type = annotation[item.name]
            if "{default}" in item.markdown and item.name in annotation.defaults:
                default = annotation.defaults[item.name]
                item.markdown = item.markdown.replace("{default}", default)

    if doc["Attributes"] is not None and annotation.attributes:
        for item in doc["Attributes"]:
            if item.type == "" and item.name in annotation.attributes:
                item.type = annotation.attributes[item.name]

    for name in ["Returns", "Yields"]:
        if doc[name] is not None:
            section = doc[name]
            if section.type == "":
                section.type = getattr(annotation, name.lower())


def parse_docstring(obj: Any) -> Docstring:
    doc = inspect.getdoc(obj)
    if not doc:
        return Docstring(obj, [])
    sections = []
    for section in split_section(doc):
        sections.append(create_section(*section))
    docstring = Docstring(obj, sections)
    postprocess(docstring, obj)
    return docstring



from typing import Iterator


def convert(text: str) -> str:
    blocks = []
    for block in split(text):
        if block.startswith(">>>"):
            block = f"~~~python\n{block}\n~~~\n"
        blocks.append(block)
    return "\n".join(blocks).strip() + "\n"


def delete_indent(lines, start, stop):
    from mkapi.core.docstring import get_indent

    indent = get_indent(lines[start])
    return "\n".join(x[indent:] for x in lines[start:stop]).strip()


def split(text: str) -> Iterator[str]:
    start = 0
    in_code = False
    lines = text.split("\n")
    for stop, line in enumerate(lines, 1):
        if ">>>" in line and not in_code:
            if start < stop - 1:
                yield "\n".join(lines[start : stop - 1])
            start = stop - 1
            in_code = True
        elif not line.strip() and in_code:
            yield delete_indent(lines, start, stop)
            start = stop
            in_code = False
    if start < len(lines):
        yield delete_indent(lines, start, len(lines))



import os
import sys

import click

from mkapi import __version__

pgk_dir = os.path.dirname(os.path.abspath(__file__))
version_msg = f"{__version__} from {pgk_dir} (Python {sys.version[:3]})."


@click.group(invoke_without_command=True)
@click.pass_context
@click.version_option(version_msg, "-V", "--version")
def cli(ctx):
    pass



import os
import shutil


def collect(top):
    base = os.path.basename(top)
    for root, dirs, files in os.walk(top):
        for file in files:
            if file.endswith(".py") and not file.startswith("__"):
                middle = root[len(top) :].replace("/", ".").replace("\\", ".")
                yield ".".join([base + middle, file[:-3]])
        for x in dirs:
            if x.startswith("__"):
                dirs.remove(x)


def make_pages(root: str, api_dir: str):
    if os.path.exists(api_dir):
        shutil.rmtree(api_dir)
    os.mkdir(api_dir)
    api_base = os.path.basename(api_dir)
    nav = []
    for module in collect(root):
        markdown = f"[mkapi]({module})\n"
        path = os.path.join(api_dir, module + ".md")
        with open(path, "w") as f:
            f.write(markdown)
        nav.append({module: "/".join([api_base, module + ".md"])})
    return nav



import importlib
import inspect
from dataclasses import dataclass, field
from functools import partial
from typing import Any, List, Optional, Tuple

import mkapi.core.preprocess
from mkapi.core.docstring import Docstring, parse_docstring
from mkapi.core.inspect import Signature

ISFUNCTIONS = {}
for x in dir(inspect):
    if x.startswith("is"):
        name = x[2:]
        if name not in ["routine", "builtin", "code"]:
            ISFUNCTIONS[name] = getattr(inspect, x)


def get_kind(kinds: List[str]) -> str:
    if "generatorfunction" in kinds:
        return "generator"
    return kinds[-1]


@dataclass
class Node:
    obj: Any = field(repr=False)
    name: str
    depth: int
    prefix: str
    kinds: List[str]
    lineno: int
    signature: Optional[Signature]
    docstring: Docstring
    members: List["Node"]

    def __post_init__(self):
        self.kind = get_kind(self.kinds)
        if self.name.startswith("__"):
            self.type = "special"
        elif self.name.startswith("_"):
            self.type = "private"
        else:
            self.type = "normal"

    def __iter__(self):
        yield from self.docstring
        for member in self.members:
            yield from member

    def get_markdown(self):
        markdowns = []
        for x in self:
            markdowns.append(mkapi.core.preprocess.convert(x.markdown))
        return "\n\n<!-- mkapi:sep -->\n\n".join(markdowns)

    def set_html(self, html: str):
        for x, html in zip(self, html.split("<!-- mkapi:sep -->")):
            x.set_html(html.strip())


def get_kinds(obj) -> List[str]:
    kinds = []
    if hasattr(obj, "__dataclass_fields__"):
        return ["dataclass"]
    for kind, func in ISFUNCTIONS.items():
        if func(obj):
            kinds.append(kind)
    if "function" in kinds and "generatorfunction" not in kinds:
        try:
            if "self" in inspect.signature(obj).parameters:
                kinds = ["method"]
        except ValueError:
            kinds = []
    if isinstance(obj, property):
        if obj.fset:
            kinds.append("readwrite_property")
        else:
            kinds.append("readonly_property")
    return kinds


def get_sourcefile_and_lineno(obj) -> Tuple[str, int]:
    if isinstance(obj, property):
        obj = obj.fget
    sourcefile = inspect.getsourcefile(obj) or ""
    lineno = inspect.getsourcelines(obj)[1]
    return sourcefile, lineno


def filter(obj, sourcefile, lineno, qualname) -> bool:
    if isinstance(obj, property):
        return True
    kinds = get_kinds(obj)
    if not qualname and 'dataclass' in kinds:
        return True
    if not kinds:
        return False

    try:
        sourcefile_, lineno_ = get_sourcefile_and_lineno(obj)
    except Exception:
        return False
    if not hasattr(obj, "__qualname__"):
        return False
    if qualname and obj.__qualname__.startswith(qualname):
        return True
    if sourcefile_ is None:
        return False
    elif sourcefile_ == sourcefile and lineno_ > lineno:
        return True
    else:
        return False


def ignore_name(name: str) -> bool:
    if name == '__init__':
        return False
    if name.startswith("_"):
        return True
    return False


def walk(name, obj, prefix="", depth=0) -> Node:
    kinds = get_kinds(obj)
    sourcefile, lineno = get_sourcefile_and_lineno(obj)
    if sourcefile.endswith("__init__.py"):
        kinds = ["package"]
    docstring = parse_docstring(obj)
    if prefix:
        next_prefix = ".".join([prefix, name])
    else:
        next_prefix = name
    if hasattr(obj, "__qualname__"):
        qualname = obj.__qualname__
    else:
        qualname = ""
    members = []
    if not isinstance(obj, property):
        func = partial(filter, sourcefile=sourcefile, lineno=lineno, qualname=qualname)
        for x in inspect.getmembers(obj, func):
            if not ignore_name(x[0]):
                member = walk(*x, prefix=next_prefix, depth=depth + 1)
                if member.type == "normal" or member.docstring:
                    members.append(member)
    if callable(obj):
        signature = Signature(obj)
    else:
        signature = None  # type:ignore

    if "class" in kinds:
        if not len(docstring):
            for member in members:
                if member.name == "__init__" and len(member.docstring):
                    markdown = member.docstring.sections[0].markdown
                    if not markdown.startswith("Initialize self"):
                        docstring = member.docstring
        members = [member for member in members if member.name != "__init__"]

    node = Node(obj, name, depth, prefix, kinds, lineno, signature, docstring, members)
    if isinstance(obj, property):
        if docstring.sections:
            node.type = docstring.sections[0].type
        else:
            node.type = ""
    return node


def get_attr(path: str):
    module_path, _, name = path.rpartition(".")
    module = importlib.import_module(module_path)
    return getattr(module, name)


def get_object(name: str):
    try:
        return get_attr(name)
    except (ModuleNotFoundError, AttributeError, ValueError):
        return importlib.import_module(name)


def get_node(name: str) -> Node:
    obj = get_object(name)
    node = walk(name, obj)
    return node



import re
from dataclasses import dataclass, field
from typing import Iterator, List

from mkapi.core.node import Node, get_node
from mkapi.core.renderer import Renderer

MKAPI_PATTERN = re.compile(r"^!\[mkapi\]\((.*?)\)$", re.MULTILINE)

HTML_PATTERN = re.compile(
    r"<!-- mkapi:(\d+):begin -->(.*?)<!-- mkapi:end -->", re.MULTILINE | re.DOTALL
)


renderer = Renderer()


@dataclass
class Page:
    source: str
    nodes: List[Node] = field(default_factory=list, init=False)

    def __post_init__(self):
        self.markdown = "\n\n".join(self.split(self.source))

    def split(self, source) -> Iterator[str]:
        cursor = 0
        index = 0
        for match in MKAPI_PATTERN.finditer(source):
            start, end = match.start(), match.end()
            if cursor < start:
                markdown = source[cursor:start].strip()
                if markdown:
                    yield markdown
            name = match.group(1)
            node = get_node(name)
            self.nodes.append(node)
            markdown = node.get_markdown()
            yield f"<!-- mkapi:{index}:begin -->\n\n{markdown}\n\n<!-- mkapi:end -->"
            cursor = end
            index += 1
        if cursor < len(source):
            markdown = source[cursor:].strip()
            if markdown:
                yield markdown

    def content(self, html):
        return re.sub(HTML_PATTERN, self.replace, html)

    def replace(self, match):
        node = self.nodes[int(match.group(1))]
        node.set_html(match.group(2))
        return renderer.render(node)
